@@ -1,25 +1,19 @@
-# -*- coding: utf-8 -*-
-from __future__ import print_function, unicode_literals
-
 import copy
+import datetime
 import locale
 import logging
 import os
 import re
-import sys
+from html import unescape
+from urllib.parse import unquote, urljoin, urlparse, urlunparse
 
 import pytz
 
-import six
-from six.moves.urllib.parse import urljoin, urlparse, urlunparse
-
-from pelican import signals
+from pelican.plugins import signals
 from pelican.settings import DEFAULT_CONFIG
-from pelican.utils import (SafeDatetime, deprecated_attribute, memoized,
-                           path_to_url, posixize_path,
-                           python_2_unicode_compatible, sanitised_join,
-                           set_date_tzinfo, slugify, strftime,
-                           truncate_html_words)
+from pelican.utils import (deprecated_attribute, memoized, path_to_url,
+                           posixize_path, sanitised_join, set_date_tzinfo,
+                           slugify, truncate_html_words)
 
 # Import these so that they're avalaible when you import from pelican.contents.
 from pelican.urlwrappers import (Author, Category, Tag, URLWrapper)  # NOQA
@@ -27,8 +21,7 @@ from pelican.urlwrappers import (Author, Category, Tag, URLWrapper)  # NOQA
 logger = logging.getLogger(__name__)
 
 
-@python_2_unicode_compatible
-class Content(object):
+class Content:
     """Represents a content.
 
     :param content: the string to parse, containing the original content.
@@ -98,18 +91,21 @@ class Content(object):
         if not hasattr(self, 'slug'):
             if (settings['SLUGIFY_SOURCE'] == 'title' and
                     hasattr(self, 'title')):
-                self.slug = slugify(
-                    self.title,
-                    regex_subs=settings.get('SLUG_REGEX_SUBSTITUTIONS', []))
+                value = self.title
             elif (settings['SLUGIFY_SOURCE'] == 'basename' and
                     source_path is not None):
-                basename = os.path.basename(
-                    os.path.splitext(source_path)[0])
+                value = os.path.basename(os.path.splitext(source_path)[0])
+            else:
+                value = None
+            if value is not None:
                 self.slug = slugify(
-                    basename,
-                    regex_subs=settings.get('SLUG_REGEX_SUBSTITUTIONS', []))
+                    value,
+                    regex_subs=settings.get('SLUG_REGEX_SUBSTITUTIONS', []),
+                    preserve_case=settings.get('SLUGIFY_PRESERVE_CASE', False),
+                    use_unicode=settings.get('SLUGIFY_USE_UNICODE', False))
 
         self.source_path = source_path
+        self.relative_source_path = self.get_relative_source_path()
 
         # manage the date format
         if not hasattr(self, 'date_format'):
@@ -120,27 +116,26 @@ class Content(object):
 
         if isinstance(self.date_format, tuple):
             locale_string = self.date_format[0]
-            if sys.version_info < (3, ) and isinstance(locale_string,
-                                                       six.text_type):
-                locale_string = locale_string.encode('ascii')
             locale.setlocale(locale.LC_ALL, locale_string)
             self.date_format = self.date_format[1]
 
         # manage timezone
         default_timezone = settings.get('TIMEZONE', 'UTC')
         timezone = getattr(self, 'timezone', default_timezone)
+        self.timezone = pytz.timezone(timezone)
 
         if hasattr(self, 'date'):
             self.date = set_date_tzinfo(self.date, timezone)
-            self.locale_date = strftime(self.date, self.date_format)
+            self.locale_date = self.date.strftime(self.date_format)
 
         if hasattr(self, 'modified'):
             self.modified = set_date_tzinfo(self.modified, timezone)
-            self.locale_modified = strftime(self.modified, self.date_format)
+            self.locale_modified = self.modified.strftime(self.date_format)
 
         # manage status
         if not hasattr(self, 'status'):
-            self.status = getattr(self, 'default_status', None)
+            # Previous default of None broke comment plugins and perhaps others
+            self.status = getattr(self, 'default_status', '')
 
         # store the summary metadata if it is set
         if 'summary' in metadata:
@@ -186,9 +181,9 @@ class Content(object):
         if hasattr(self, 'allowed_statuses'):
             if self.status not in self.allowed_statuses:
                 logger.error(
-                    "Unknown status '%s' for file %s, skipping it.",
+                    "Unknown status '%s' for file %s, skipping it. (Not in %s)",
                     self.status,
-                    self
+                    self, self.allowed_statuses
                 )
                 return False
 
@@ -211,9 +206,8 @@ class Content(object):
             'path': path_to_url(path),
             'slug': getattr(self, 'slug', ''),
             'lang': getattr(self, 'lang', 'en'),
-            'date': getattr(self, 'date', SafeDatetime.now()),
+            'date': getattr(self, 'date', datetime.datetime.now()),
             'author': self.author.slug if hasattr(self, 'author') else '',
-            'tag': self.tag.slug if hasattr(self, 'tag') else '',
             'category': self.category.slug if hasattr(self, 'category') else ''
         })
         return metadata
@@ -221,7 +215,7 @@ class Content(object):
     def _expand_settings(self, key, klass=None):
         if not klass:
             klass = self.__class__.__name__
-        fq_key = ('%s_%s' % (klass, key)).upper()
+        fq_key = ('{}_{}'.format(klass, key)).upper()
         return self.settings[fq_key].format(**self.url_format)
 
     def get_url_setting(self, key):
@@ -257,38 +251,55 @@ class Content(object):
 
         # XXX Put this in a different location.
         if what in {'filename', 'static', 'attach'}:
-            if path.startswith('/'):
-                path = path[1:]
+            def _get_linked_content(key, url):
+                nonlocal value
+
+                def _find_path(path):
+                    if path.startswith('/'):
+                        path = path[1:]
+                    else:
+                        # relative to the source path of this content
+                        path = self.get_relative_source_path(
+                            os.path.join(self.relative_dir, path)
+                        )
+                    return self._context[key].get(path, None)
+
+                # try path
+                result = _find_path(url.path)
+                if result is not None:
+                    return result
+
+                # try unquoted path
+                result = _find_path(unquote(url.path))
+                if result is not None:
+                    return result
+
+                # try html unescaped url
+                unescaped_url = urlparse(unescape(url.geturl()))
+                result = _find_path(unescaped_url.path)
+                if result is not None:
+                    value = unescaped_url
+                    return result
+
+                # check if a static file is linked with {filename}
+                if what == 'filename' and key == 'generated_content':
+                    linked_content = _get_linked_content('static_content', value)
+                    if linked_content:
+                        logger.warning(
+                            '{filename} used for linking to static'
+                            ' content %s in %s. Use {static} instead',
+                            value.path,
+                            self.get_relative_source_path())
+                        return linked_content
+
+                return None
+
+            if what == 'filename':
+                key = 'generated_content'
             else:
-                # relative to the source path of this content
-                path = self.get_relative_source_path(
-                    os.path.join(self.relative_dir, path)
-                )
+                key = 'static_content'
 
-            key = 'static_content' if what in ('static', 'attach')\
-                else 'generated_content'
-
-            def _get_linked_content(key, path):
-                try:
-                    return self._context[key][path]
-                except KeyError:
-                    try:
-                        # Markdown escapes spaces, try unescaping
-                        return self._context[key][path.replace('%20', ' ')]
-                    except KeyError:
-                        if what == 'filename' and key == 'generated_content':
-                            key = 'static_content'
-                            linked_content = _get_linked_content(key, path)
-                            if linked_content:
-                                logger.warning(
-                                    '{filename} used for linking to static'
-                                    'content %s in %s. Use {static} instead',
-                                    path,
-                                    self.get_relative_source_path())
-                                return linked_content
-                        return None
-
-            linked_content = _get_linked_content(key, path)
+            linked_content = _get_linked_content(key, value)
             if linked_content:
                 if what == 'attach':
                     linked_content.attach_to(self)
@@ -329,7 +340,7 @@ class Content(object):
                 (?:href|src|poster|data|cite|formaction|action)\s*=\s*)
 
             (?P<quote>["\'])      # require value to be quoted
-            (?P<path>{0}(?P<value>.*?))  # the url value
+            (?P<path>{}(?P<value>.*?))  # the url value
             \2""".format(intrasite_link_regex)
         return re.compile(regex, re.X)
 
@@ -398,7 +409,8 @@ class Content(object):
             return self.content
 
         return truncate_html_words(self.content,
-                                   self.settings['SUMMARY_MAX_LENGTH'])
+                                   self.settings['SUMMARY_MAX_LENGTH'],
+                                   self.settings['SUMMARY_END_SUFFIX'])
 
     @property
     def summary(self):
@@ -496,40 +508,43 @@ class Page(Content):
 
     def _expand_settings(self, key):
         klass = 'draft_page' if self.status == 'draft' else None
-        return super(Page, self)._expand_settings(key, klass)
+        return super()._expand_settings(key, klass)
 
 
 class Article(Content):
     mandatory_properties = ('title', 'date', 'category')
-    allowed_statuses = ('published', 'draft')
+    allowed_statuses = ('published', 'hidden', 'draft')
     default_status = 'published'
     default_template = 'article'
 
     def __init__(self, *args, **kwargs):
-        super(Article, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
         # handle WITH_FUTURE_DATES (designate article to draft based on date)
         if not self.settings['WITH_FUTURE_DATES'] and hasattr(self, 'date'):
             if self.date.tzinfo is None:
-                now = SafeDatetime.now()
+                now = datetime.datetime.now()
             else:
-                now = SafeDatetime.utcnow().replace(tzinfo=pytz.utc)
+                now = datetime.datetime.utcnow().replace(tzinfo=pytz.utc)
             if self.date > now:
                 self.status = 'draft'
 
         # if we are a draft and there is no date provided, set max datetime
         if not hasattr(self, 'date') and self.status == 'draft':
-            self.date = SafeDatetime.max
+            self.date = datetime.datetime.max.replace(tzinfo=self.timezone)
 
     def _expand_settings(self, key):
         klass = 'draft' if self.status == 'draft' else 'article'
-        return super(Article, self)._expand_settings(key, klass)
+        return super()._expand_settings(key, klass)
 
 
-@python_2_unicode_compatible
-class Static(Page):
+class Static(Content):
+    mandatory_properties = ('title',)
+    default_status = 'published'
+    default_template = None
+
     def __init__(self, *args, **kwargs):
-        super(Static, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self._output_location_referenced = False
 
     @deprecated_attribute(old='filepath', new='source_path', since=(3, 2, 0))
@@ -548,13 +563,13 @@ class Static(Page):
     def url(self):
         # Note when url has been referenced, so we can avoid overriding it.
         self._output_location_referenced = True
-        return super(Static, self).url
+        return super().url
 
     @property
     def save_as(self):
         # Note when save_as has been referenced, so we can avoid overriding it.
         self._output_location_referenced = True
-        return super(Static, self).save_as
+        return super().save_as
 
     def attach_to(self, content):
         """Override our output directory with that of the given content object.
