@@ -1,50 +1,54 @@
-# -*- coding: utf-8 -*-
-from __future__ import print_function, unicode_literals
-
 import argparse
-import collections
-import locale
+import json
 import logging
 import multiprocessing
 import os
 import pprint
-import re
 import sys
 import time
 import traceback
-
-import six
+from collections.abc import Iterable
+# Combines all paths to `pelican` package accessible from `sys.path`
+# Makes it possible to install `pelican` and namespace plugins into different
+# locations in the file system (e.g. pip with `-e` or `--user`)
+from pkgutil import extend_path
+__path__ = extend_path(__path__, __name__)
 
 # pelican.log has to be the first pelican module to be loaded
 # because logging.setLoggerClass has to be called before logging.getLogger
+from pelican.log import console
 from pelican.log import init as init_logging
-from pelican import signals  # noqa
-from pelican.generators import (ArticlesGenerator, PagesGenerator,
-                                SourceFileGenerator, StaticGenerator,
-                                TemplatePagesGenerator)
+from pelican.generators import (ArticlesGenerator,  # noqa: I100
+                                PagesGenerator, SourceFileGenerator,
+                                StaticGenerator, TemplatePagesGenerator)
+from pelican.plugins import signals
+from pelican.plugins._utils import get_plugin_name, load_plugins
 from pelican.readers import Readers
 from pelican.server import ComplexHTTPRequestHandler, RootedHTTPServer
 from pelican.settings import read_settings
-from pelican.utils import (clean_output_dir, file_watcher,
-                           folder_watcher, maybe_pluralize)
+from pelican.utils import (FileSystemWatcher, clean_output_dir, maybe_pluralize)
 from pelican.writers import Writer
 
-__version__ = "4.0.0"
+try:
+    __version__ = __import__('pkg_resources') \
+        .get_distribution('pelican').version
+except Exception:
+    __version__ = "unknown"
+
 DEFAULT_CONFIG_NAME = 'pelicanconf.py'
 logger = logging.getLogger(__name__)
 
 
-class Pelican(object):
+class Pelican:
 
     def __init__(self, settings):
-        """Pelican initialisation
+        """Pelican initialization
 
         Performs some checks on the environment before doing anything else.
         """
 
         # define the default settings
         self.settings = settings
-        self._handle_deprecation()
 
         self.path = settings['PATH']
         self.theme = settings['THEME']
@@ -64,86 +68,17 @@ class Pelican(object):
 
     def init_plugins(self):
         self.plugins = []
-        logger.debug('Temporarily adding PLUGIN_PATHS to system path')
-        _sys_path = sys.path[:]
-        for pluginpath in self.settings['PLUGIN_PATHS']:
-            sys.path.insert(0, pluginpath)
-        for plugin in self.settings['PLUGINS']:
-            # if it's a string, then import it
-            if isinstance(plugin, six.string_types):
-                logger.debug("Loading plugin `%s`", plugin)
-                try:
-                    plugin = __import__(plugin, globals(), locals(),
-                                        str('module'))
-                except ImportError as e:
-                    logger.error(
-                        "Cannot load plugin `%s`\n%s", plugin, e)
-                    continue
+        for plugin in load_plugins(self.settings):
+            name = get_plugin_name(plugin)
+            logger.debug('Registering plugin `%s`', name)
+            try:
+                plugin.register()
+                self.plugins.append(plugin)
+            except Exception as e:
+                logger.error('Cannot register plugin `%s`\n%s',
+                             name, e)
 
-            logger.debug("Registering plugin `%s`", plugin.__name__)
-            plugin.register()
-            self.plugins.append(plugin)
-        logger.debug('Restoring system path')
-        sys.path = _sys_path
-
-    def _handle_deprecation(self):
-
-        if self.settings.get('CLEAN_URLS', False):
-            logger.warning('Found deprecated `CLEAN_URLS` in settings.'
-                           ' Modifying the following settings for the'
-                           ' same behaviour.')
-
-            self.settings['ARTICLE_URL'] = '{slug}/'
-            self.settings['ARTICLE_LANG_URL'] = '{slug}-{lang}/'
-            self.settings['PAGE_URL'] = 'pages/{slug}/'
-            self.settings['PAGE_LANG_URL'] = 'pages/{slug}-{lang}/'
-
-            for setting in ('ARTICLE_URL', 'ARTICLE_LANG_URL', 'PAGE_URL',
-                            'PAGE_LANG_URL'):
-                logger.warning("%s = '%s'", setting, self.settings[setting])
-
-        if self.settings.get('AUTORELOAD_IGNORE_CACHE'):
-            logger.warning('Found deprecated `AUTORELOAD_IGNORE_CACHE` in '
-                           'settings. Use --ignore-cache instead.')
-            self.settings.pop('AUTORELOAD_IGNORE_CACHE')
-
-        if self.settings.get('ARTICLE_PERMALINK_STRUCTURE', False):
-            logger.warning('Found deprecated `ARTICLE_PERMALINK_STRUCTURE` in'
-                           ' settings.  Modifying the following settings for'
-                           ' the same behaviour.')
-
-            structure = self.settings['ARTICLE_PERMALINK_STRUCTURE']
-
-            # Convert %(variable) into {variable}.
-            structure = re.sub(r'%\((\w+)\)s', r'{\g<1>}', structure)
-
-            # Convert %x into {date:%x} for strftime
-            structure = re.sub(r'(%[A-z])', r'{date:\g<1>}', structure)
-
-            # Strip a / prefix
-            structure = re.sub('^/', '', structure)
-
-            for setting in ('ARTICLE_URL', 'ARTICLE_LANG_URL', 'PAGE_URL',
-                            'PAGE_LANG_URL', 'DRAFT_URL', 'DRAFT_LANG_URL',
-                            'ARTICLE_SAVE_AS', 'ARTICLE_LANG_SAVE_AS',
-                            'DRAFT_SAVE_AS', 'DRAFT_LANG_SAVE_AS',
-                            'PAGE_SAVE_AS', 'PAGE_LANG_SAVE_AS'):
-                self.settings[setting] = os.path.join(structure,
-                                                      self.settings[setting])
-                logger.warning("%s = '%s'", setting, self.settings[setting])
-
-        for new, old in [('FEED', 'FEED_ATOM'), ('TAG_FEED', 'TAG_FEED_ATOM'),
-                         ('CATEGORY_FEED', 'CATEGORY_FEED_ATOM'),
-                         ('TRANSLATION_FEED', 'TRANSLATION_FEED_ATOM')]:
-            if self.settings.get(new, False):
-                logger.warning(
-                    'Found deprecated `%(new)s` in settings. Modify %(new)s '
-                    'to %(old)s in your settings and theme for the same '
-                    'behavior. Temporarily setting %(old)s for backwards '
-                    'compatibility.',
-                    {'new': new, 'old': old}
-                )
-                self.settings[old] = self.settings[new]
+        self.settings['PLUGINS'] = [get_plugin_name(p) for p in self.plugins]
 
     def run(self):
         """Run the generators and return"""
@@ -164,13 +99,15 @@ class Pelican(object):
                 path=self.path,
                 theme=self.theme,
                 output_path=self.output_path,
-            ) for cls in self.get_generator_classes()
+            ) for cls in self._get_generator_classes()
         ]
 
-        # erase the directory if it is not the source and if that's
-        # explicitly asked
-        if (self.delete_outputdir and not
-                os.path.realpath(self.path).startswith(self.output_path)):
+        # Delete the output directory if (1) the appropriate setting is True
+        # and (2) that directory is not the parent of the source directory
+        if (self.delete_outputdir
+                and os.path.commonpath([os.path.realpath(self.output_path)]) !=
+                os.path.commonpath([os.path.realpath(self.output_path),
+                                    os.path.realpath(self.path)])):
             clean_output_dir(self.output_path, self.output_retention)
 
         for p in generators:
@@ -183,7 +120,7 @@ class Pelican(object):
 
         signals.all_generators_finalized.send(generators)
 
-        writer = self.get_writer()
+        writer = self._get_writer()
 
         for p in generators:
             if hasattr(p, 'generate_output'):
@@ -206,6 +143,11 @@ class Pelican(object):
              len(articles_generator.drafts_translations)),
             'draft',
             'drafts')
+        pluralized_hidden_articles = maybe_pluralize(
+            (len(articles_generator.hidden_articles) +
+             len(articles_generator.hidden_translations)),
+            'hidden article',
+            'hidden articles')
         pluralized_pages = maybe_pluralize(
             (len(pages_generator.pages) +
              len(pages_generator.translations)),
@@ -222,84 +164,128 @@ class Pelican(object):
             'draft page',
             'draft pages')
 
-        print('Done: Processed {}, {}, {}, {} and {} in {:.2f} seconds.'
-              .format(
-                    pluralized_articles,
-                    pluralized_drafts,
-                    pluralized_pages,
-                    pluralized_hidden_pages,
-                    pluralized_draft_pages,
-                    time.time() - start_time))
+        console.print('Done: Processed {}, {}, {}, {}, {} and {} in {:.2f} seconds.'
+                      .format(
+                              pluralized_articles,
+                              pluralized_drafts,
+                              pluralized_hidden_articles,
+                              pluralized_pages,
+                              pluralized_hidden_pages,
+                              pluralized_draft_pages,
+                              time.time() - start_time))
 
-    def get_generator_classes(self):
-        generators = [ArticlesGenerator, PagesGenerator]
+    def _get_generator_classes(self):
+        discovered_generators = [
+            (ArticlesGenerator, "internal"),
+            (PagesGenerator, "internal")
+        ]
 
-        if self.settings['TEMPLATE_PAGES']:
-            generators.append(TemplatePagesGenerator)
-        if self.settings['OUTPUT_SOURCES']:
-            generators.append(SourceFileGenerator)
+        if self.settings["TEMPLATE_PAGES"]:
+            discovered_generators.append((TemplatePagesGenerator, "internal"))
 
-        for pair in signals.get_generators.send(self):
-            (funct, value) = pair
+        if self.settings["OUTPUT_SOURCES"]:
+            discovered_generators.append((SourceFileGenerator, "internal"))
 
-            if not isinstance(value, collections.Iterable):
-                value = (value, )
-
-            for v in value:
-                if isinstance(v, type):
-                    logger.debug('Found generator: %s', v)
-                    generators.append(v)
+        for receiver, values in signals.get_generators.send(self):
+            if not isinstance(values, Iterable):
+                values = (values,)
+            for generator in values:
+                if generator is None:
+                    continue  # plugin did not return a generator
+                discovered_generators.append((generator, receiver.__module__))
 
         # StaticGenerator must run last, so it can identify files that
         # were skipped by the other generators, and so static files can
         # have their output paths overridden by the {attach} link syntax.
-        generators.append(StaticGenerator)
+        discovered_generators.append((StaticGenerator, "internal"))
+
+        generators = []
+
+        for generator, origin in discovered_generators:
+            if not isinstance(generator, type):
+                logger.error("Generator %s (%s) cannot be loaded", generator, origin)
+                continue
+
+            logger.debug("Found generator: %s (%s)", generator.__name__, origin)
+            generators.append(generator)
+
         return generators
 
-    def get_writer(self):
-        writers = [w for (_, w) in signals.get_writer.send(self)
-                   if isinstance(w, type)]
-        writers_found = len(writers)
-        if writers_found == 0:
+    def _get_writer(self):
+        writers = [w for _, w in signals.get_writer.send(self) if isinstance(w, type)]
+        num_writers = len(writers)
+
+        if num_writers == 0:
             return Writer(self.output_path, settings=self.settings)
-        else:
-            writer = writers[0]
-            if writers_found == 1:
-                logger.debug('Found writer: %s', writer)
-            else:
-                logger.warning(
-                    '%s writers found, using only first one: %s',
-                    writers_found, writer)
-            return writer(self.output_path, settings=self.settings)
+
+        if num_writers > 1:
+            logger.warning("%s writers found, using only first one", num_writers)
+
+        writer = writers[0]
+
+        logger.debug("Found writer: %s (%s)", writer.__name__, writer.__module__)
+        return writer(self.output_path, settings=self.settings)
 
 
 class PrintSettings(argparse.Action):
     def __call__(self, parser, namespace, values, option_string):
-        instance, settings = get_instance(namespace)
+        init_logging(name=__name__)
+
+        try:
+            instance, settings = get_instance(namespace)
+        except Exception as e:
+            logger.critical("%s: %s", e.__class__.__name__, e)
+            console.print_exception()
+            sys.exit(getattr(e, 'exitcode', 1))
 
         if values:
             # One or more arguments provided, so only print those settings
             for setting in values:
                 if setting in settings:
                     # Only add newline between setting name and value if dict
-                    if isinstance(settings[setting], dict):
+                    if isinstance(settings[setting], (dict, tuple, list)):
                         setting_format = '\n{}:\n{}'
                     else:
                         setting_format = '\n{}: {}'
-                    print(setting_format.format(
+                    console.print(setting_format.format(
                         setting,
                         pprint.pformat(settings[setting])))
                 else:
-                    print('\n{} is not a recognized setting.'.format(setting))
+                    console.print('\n{} is not a recognized setting.'.format(setting))
                     break
         else:
             # No argument was given to --print-settings, so print all settings
-            pprint.pprint(settings)
+            console.print(settings)
 
         parser.exit()
 
 
-def parse_arguments():
+class ParseOverrides(argparse.Action):
+    def __call__(self, parser, namespace, values, option_string=None):
+        overrides = {}
+        for item in values:
+            try:
+                k, v = item.split("=", 1)
+            except ValueError:
+                raise ValueError(
+                    'Extra settings must be specified as KEY=VALUE pairs '
+                    f'but you specified {item}'
+                )
+            try:
+                overrides[k] = json.loads(v)
+            except json.decoder.JSONDecodeError:
+                raise ValueError(
+                    f'Invalid JSON value: {v}. '
+                    'Values specified via -e / --extra-settings flags '
+                    'must be in JSON notation. '
+                    'Use -e KEY=\'"string"\' to specify a string value; '
+                    '-e KEY=null to specify None; '
+                    '-e KEY=false (or true) to specify False (or True).'
+                )
+        setattr(namespace, self.dest, overrides)
+
+
+def parse_arguments(argv=None):
     parser = argparse.ArgumentParser(
         description='A tool to generate a static blog, '
                     ' with restructured text input files.',
@@ -322,7 +308,7 @@ def parse_arguments():
 
     parser.add_argument('-s', '--settings', dest='settings',
                         help='The settings of the application, this is '
-                        'automatically set to {0} if a file exists with this '
+                        'automatically set to {} if a file exists with this '
                         'name.'.format(DEFAULT_CONFIG_NAME))
 
     parser.add_argument('-d', '--delete-output-directory',
@@ -392,7 +378,17 @@ def parse_arguments():
                         help='IP to bind to when serving files via HTTP '
                         '(default: 127.0.0.1)')
 
-    args = parser.parse_args()
+    parser.add_argument('-e', '--extra-settings', dest='overrides',
+                        help='Specify one or more SETTING=VALUE pairs to '
+                             'override settings. VALUE must be in JSON notation: '
+                             'specify string values as SETTING=\'"some string"\'; '
+                             'booleans as SETTING=true or SETTING=false; '
+                             'None as SETTING=null.',
+                        nargs='*',
+                        action=ParseOverrides,
+                        default={})
+
+    args = parser.parse_args(argv)
 
     if args.port is not None and not args.listen:
         logger.warning('--port without --listen has no effect')
@@ -403,6 +399,8 @@ def parse_arguments():
 
 
 def get_config(args):
+    """Builds a config dictionary based on supplied `args`.
+    """
     config = {}
     if args.path:
         config['PATH'] = os.path.abspath(os.path.expanduser(args.path))
@@ -427,16 +425,8 @@ def get_config(args):
     if args.bind is not None:
         config['BIND'] = args.bind
     config['DEBUG'] = args.verbosity == logging.DEBUG
+    config.update(args.overrides)
 
-    # argparse returns bytes in Py2. There is no definite answer as to which
-    # encoding argparse (or sys.argv) uses.
-    # "Best" option seems to be locale.getpreferredencoding()
-    # http://mail.python.org/pipermail/python-list/2006-October/405766.html
-    if not six.PY3:
-        enc = locale.getpreferredencoding()
-        for key in config:
-            if key in ('PATH', 'OUTPUT_PATH', 'THEME'):
-                config[key] = config[key].decode(enc)
     return config
 
 
@@ -444,13 +434,13 @@ def get_instance(args):
 
     config_file = args.settings
     if config_file is None and os.path.isfile(DEFAULT_CONFIG_NAME):
-            config_file = DEFAULT_CONFIG_NAME
-            args.settings = DEFAULT_CONFIG_NAME
+        config_file = DEFAULT_CONFIG_NAME
+        args.settings = DEFAULT_CONFIG_NAME
 
     settings = read_settings(config_file, override=get_config(args))
 
     cls = settings['PELICAN_CLASS']
-    if isinstance(cls, six.string_types):
+    if isinstance(cls, str):
         module, cls_name = cls.rsplit('.', 1)
         module = __import__(module)
         cls = getattr(module, cls_name)
@@ -458,65 +448,36 @@ def get_instance(args):
     return cls(settings), settings
 
 
-def autoreload(watchers, args, old_static, reader_descs, excqueue=None):
+def autoreload(args, excqueue=None):
+    console.print('  --- AutoReload Mode: Monitoring `content`, `theme` and'
+                  ' `settings` for changes. ---')
+    pelican, settings = get_instance(args)
+    watcher = FileSystemWatcher(args.settings, Readers, settings)
+    sleep = False
     while True:
         try:
-            # Check source dir for changed files ending with the given
-            # extension in the settings. In the theme dir is no such
-            # restriction; all files are recursively checked if they
-            # have changed, no matter what extension the filenames
-            # have.
-            modified = {k: next(v) for k, v in watchers.items()}
+            # Don't sleep first time, but sleep afterwards to reduce cpu load
+            if sleep:
+                time.sleep(0.5)
+            else:
+                sleep = True
+
+            modified = watcher.check()
 
             if modified['settings']:
                 pelican, settings = get_instance(args)
-
-                # Adjust static watchers if there are any changes
-                new_static = settings.get("STATIC_PATHS", [])
-
-                # Added static paths
-                # Add new watchers and set them as modified
-                new_watchers = set(new_static).difference(old_static)
-                for static_path in new_watchers:
-                    static_key = '[static]%s' % static_path
-                    watchers[static_key] = folder_watcher(
-                        os.path.join(pelican.path, static_path),
-                        [''],
-                        pelican.ignore_files)
-                    modified[static_key] = next(watchers[static_key])
-
-                # Removed static paths
-                # Remove watchers and modified values
-                old_watchers = set(old_static).difference(new_static)
-                for static_path in old_watchers:
-                    static_key = '[static]%s' % static_path
-                    watchers.pop(static_key)
-                    modified.pop(static_key)
-
-                # Replace old_static with the new one
-                old_static = new_static
+                watcher.update_watchers(settings)
 
             if any(modified.values()):
-                print('\n-> Modified: {}. re-generating...'.format(
-                    ', '.join(k for k, v in modified.items() if v)))
-
-                if modified['content'] is None:
-                    logger.warning(
-                        'No valid files found in content for '
-                        + 'the active readers:\n'
-                        + '\n'.join(reader_descs))
-
-                if modified['theme'] is None:
-                    logger.warning('Empty theme folder. Using `basic` '
-                                   'theme.')
-
+                console.print('\n-> Modified: {}. re-generating...'.format(
+                              ', '.join(k for k, v in modified.items() if v)))
                 pelican.run()
 
-        except KeyboardInterrupt as e:
-            logger.warning("Keyboard interrupt, quitting.")
+        except KeyboardInterrupt:
             if excqueue is not None:
-                excqueue.put(traceback.format_exception_only(type(e), e)[-1])
-            return
+                excqueue.put(None)
+                return
+            raise
 
         except Exception as e:
             if (args.verbosity == logging.DEBUG):
@@ -526,13 +487,15 @@ def autoreload(watchers, args, old_static, reader_descs, excqueue=None):
                 else:
                     raise
             logger.warning(
-                'Caught exception "%s". Reloading.', e)
-
-        finally:
-            time.sleep(.5)  # sleep to avoid cpu load
+                'Caught exception:\n"%s".', e,
+                exc_info=settings.get('DEBUG', False))
 
 
 def listen(server, port, output, excqueue=None):
+    # set logging level to at least "INFO" (so we can see the server requests)
+    if logger.level < logging.INFO:
+        logger.setLevel(logging.INFO)
+
     RootedHTTPServer.allow_reuse_address = True
     try:
         httpd = RootedHTTPServer(
@@ -543,20 +506,27 @@ def listen(server, port, output, excqueue=None):
             excqueue.put(traceback.format_exception_only(type(e), e)[-1])
         return
 
-    logging.info("Serving at port %s, server %s.", port, server)
     try:
+        console.print("Serving site at: http://{}:{} - Tap CTRL-C to stop".format(
+                      server, port))
         httpd.serve_forever()
     except Exception as e:
         if excqueue is not None:
             excqueue.put(traceback.format_exception_only(type(e), e)[-1])
         return
 
+    except KeyboardInterrupt:
+        httpd.socket.close()
+        if excqueue is not None:
+            return
+        raise
 
-def main():
-    args = parse_arguments()
+
+def main(argv=None):
+    args = parse_arguments(argv)
     logs_dedup_min_level = getattr(logging, args.logs_dedup_min_level)
-    init_logging(args.verbosity, args.fatal,
-                 logs_dedup_min_level=logs_dedup_min_level)
+    init_logging(level=args.verbosity, fatal=args.fatal,
+                 name=__name__, logs_dedup_min_level=logs_dedup_min_level)
 
     logger.debug('Pelican version: %s', __version__)
     logger.debug('Python version: %s', sys.version.split()[0])
@@ -564,35 +534,11 @@ def main():
     try:
         pelican, settings = get_instance(args)
 
-        readers = Readers(settings)
-        reader_descs = sorted(set(['%s (%s)' %
-                                   (type(r).__name__,
-                                    ', '.join(r.file_extensions))
-                                   for r in readers.readers.values()
-                                   if r.enabled]))
-
-        watchers = {'content': folder_watcher(pelican.path,
-                                              readers.extensions,
-                                              pelican.ignore_files),
-                    'theme': folder_watcher(pelican.theme,
-                                            [''],
-                                            pelican.ignore_files),
-                    'settings': file_watcher(args.settings)}
-
-        old_static = settings.get("STATIC_PATHS", [])
-        for static_path in old_static:
-            # use a prefix to avoid possible overriding of standard watchers
-            # above
-            watchers['[static]%s' % static_path] = folder_watcher(
-                os.path.join(pelican.path, static_path),
-                [''],
-                pelican.ignore_files)
-
         if args.autoreload and args.listen:
             excqueue = multiprocessing.Queue()
             p1 = multiprocessing.Process(
                 target=autoreload,
-                args=(watchers, args, old_static, reader_descs, excqueue))
+                args=(args, excqueue))
             p2 = multiprocessing.Process(
                 target=listen,
                 args=(settings.get('BIND'), settings.get('PORT'),
@@ -602,30 +548,23 @@ def main():
             exc = excqueue.get()
             p1.terminate()
             p2.terminate()
-            logger.critical(exc)
+            if exc is not None:
+                logger.critical(exc)
         elif args.autoreload:
-            print('  --- AutoReload Mode: Monitoring `content`, `theme` and'
-                  ' `settings` for changes. ---')
-            autoreload(watchers, args, old_static, reader_descs)
+            autoreload(args)
         elif args.listen:
             listen(settings.get('BIND'), settings.get('PORT'),
                    settings.get("OUTPUT_PATH"))
         else:
-            if next(watchers['content']) is None:
-                logger.warning(
-                    'No valid files found in content for '
-                    + 'the active readers:\n'
-                    + '\n'.join(reader_descs))
-
-            if next(watchers['theme']) is None:
-                logger.warning('Empty theme folder. Using `basic` theme.')
-
-            pelican.run()
-
+            watcher = FileSystemWatcher(args.settings, Readers, settings)
+            watcher.check()
+            with console.status("Generating..."):
+                pelican.run()
+    except KeyboardInterrupt:
+        logger.warning('Keyboard interrupt received. Exiting.')
     except Exception as e:
-        logger.critical('%s', e)
+        logger.critical("%s: %s", e.__class__.__name__, e)
 
         if args.verbosity == logging.DEBUG:
-            raise
-        else:
-            sys.exit(getattr(e, 'exitcode', 1))
+            console.print_exception()
+        sys.exit(getattr(e, 'exitcode', 1))
