@@ -1,19 +1,18 @@
-from __future__ import unicode_literals, absolute_import
-
+import datetime
 import locale
+import logging
 import os.path
-import six
-
-from jinja2.utils import generate_lorem_ipsum
+from posixpath import join as posix_join
 from sys import platform
 
-from pelican.contents import (Page, Article, Static, URLWrapper,
-                              Author, Category)
+from jinja2.utils import generate_lorem_ipsum
+
+from pelican.contents import Article, Author, Category, Page, Static
+from pelican.plugins.signals import content_object_init
 from pelican.settings import DEFAULT_CONFIG
-from pelican.signals import content_object_init
-from pelican.tests.support import unittest, get_settings
-from pelican.utils import (path_to_url, truncate_html_words, SafeDatetime,
-                           posix_join)
+from pelican.tests.support import (LoggedTestCase, get_context, get_settings,
+                                   unittest)
+from pelican.utils import (path_to_url, posixize_path, truncate_html_words)
 
 
 # generate one paragraph, enclosed with <p>
@@ -21,16 +20,19 @@ TEST_CONTENT = str(generate_lorem_ipsum(n=1))
 TEST_SUMMARY = generate_lorem_ipsum(n=1, html=False)
 
 
-class TestPage(unittest.TestCase):
+class TestBase(LoggedTestCase):
 
     def setUp(self):
-        super(TestPage, self).setUp()
+        super().setUp()
         self.old_locale = locale.setlocale(locale.LC_ALL)
-        locale.setlocale(locale.LC_ALL, str('C'))
+        locale.setlocale(locale.LC_ALL, 'C')
         self.page_kwargs = {
             'content': TEST_CONTENT,
             'context': {
                 'localsiteurl': '',
+                'generated_content': {},
+                'static_content': {},
+                'static_links': set()
             },
             'metadata': {
                 'summary': TEST_SUMMARY,
@@ -39,16 +41,40 @@ class TestPage(unittest.TestCase):
             },
             'source_path': '/path/to/file/foo.ext'
         }
+        self._disable_limit_filter()
 
     def tearDown(self):
         locale.setlocale(locale.LC_ALL, self.old_locale)
+        self._enable_limit_filter()
 
+    def _disable_limit_filter(self):
+        from pelican.contents import logger
+        logger.disable_filter()
+
+    def _enable_limit_filter(self):
+        from pelican.contents import logger
+        logger.enable_filter()
+
+    def _copy_page_kwargs(self):
+        # make a deep copy of page_kwargs
+        page_kwargs = {key: self.page_kwargs[key] for key in self.page_kwargs}
+        for key in page_kwargs:
+            if not isinstance(page_kwargs[key], dict):
+                break
+            page_kwargs[key] = {
+                subkey: page_kwargs[key][subkey] for subkey in page_kwargs[key]
+            }
+
+        return page_kwargs
+
+
+class TestPage(TestBase):
     def test_use_args(self):
         # Creating a page with arguments passed to the constructor should use
         # them to initialise object's attributes.
         metadata = {'foo': 'bar', 'foobar': 'baz', 'title': 'foobar', }
         page = Page(TEST_CONTENT, metadata=metadata,
-                context={'localsiteurl': ''})
+                    context={'localsiteurl': ''})
         for key, value in metadata.items():
             self.assertTrue(hasattr(page, key))
             self.assertEqual(value, getattr(page, key))
@@ -57,11 +83,13 @@ class TestPage(unittest.TestCase):
     def test_mandatory_properties(self):
         # If the title is not set, must throw an exception.
         page = Page('content')
-        with self.assertRaises(NameError):
-            page.check_properties()
-
+        self.assertFalse(page._has_valid_mandatory_properties())
+        self.assertLogCountEqual(
+                count=1,
+                msg="Skipping .*: could not find information about 'title'",
+                level=logging.ERROR)
         page = Page('content', metadata={'title': 'foobar'})
-        page.check_properties()
+        self.assertTrue(page._has_valid_mandatory_properties())
 
     def test_summary_from_metadata(self):
         # If a :summary: metadata is given, it should be used
@@ -85,6 +113,32 @@ class TestPage(unittest.TestCase):
         page = Page(**page_kwargs)
         self.assertEqual(page.summary, '')
 
+    def test_summary_end_suffix(self):
+        # If a :SUMMARY_END_SUFFIX: is set, and there is no other summary,
+        # generated summary should contain the specified marker at the end.
+        page_kwargs = self._copy_page_kwargs()
+        settings = get_settings()
+        page_kwargs['settings'] = settings
+        del page_kwargs['metadata']['summary']
+        settings['SUMMARY_END_SUFFIX'] = 'test_marker'
+        settings['SUMMARY_MAX_LENGTH'] = 10
+        page = Page(**page_kwargs)
+        self.assertEqual(page.summary, truncate_html_words(TEST_CONTENT, 10,
+                                                           'test_marker'))
+        self.assertIn('test_marker', page.summary)
+
+    def test_summary_get_summary_warning(self):
+        """calling ._get_summary() should issue a warning"""
+        page_kwargs = self._copy_page_kwargs()
+        page = Page(**page_kwargs)
+        self.assertEqual(page.summary, TEST_SUMMARY)
+        self.assertEqual(page._get_summary(), TEST_SUMMARY)
+        self.assertLogCountEqual(
+                count=1,
+                msg=r"_get_summary\(\) has been deprecated since 3\.6\.4\. "
+                    "Use the summary decorator instead",
+                level=logging.WARNING)
+
     def test_slug(self):
         page_kwargs = self._copy_page_kwargs()
         settings = get_settings()
@@ -96,6 +150,32 @@ class TestPage(unittest.TestCase):
         page = Page(**page_kwargs)
         self.assertEqual(page.slug, 'foo')
 
+        # test slug from title with unicode and case
+
+        inputs = (
+            # (title, expected, preserve_case, use_unicode)
+            ('指導書', 'zhi-dao-shu', False, False),
+            ('指導書', 'Zhi-Dao-Shu', True, False),
+            ('指導書', '指導書', False, True),
+            ('指導書', '指導書', True, True),
+            ('Çığ', 'cig', False, False),
+            ('Çığ', 'Cig', True, False),
+            ('Çığ', 'çığ', False, True),
+            ('Çığ', 'Çığ', True, True),
+        )
+
+        settings = get_settings()
+        page_kwargs = self._copy_page_kwargs()
+        page_kwargs['settings'] = settings
+
+        for title, expected, preserve_case, use_unicode in inputs:
+            settings['SLUGIFY_PRESERVE_CASE'] = preserve_case
+            settings['SLUGIFY_USE_UNICODE'] = use_unicode
+            page_kwargs['metadata']['title'] = title
+            page = Page(**page_kwargs)
+            self.assertEqual(page.slug, expected,
+                             (title, preserve_case, use_unicode))
+
     def test_defaultlang(self):
         # If no lang is given, default to the default one.
         page = Page(**self.page_kwargs)
@@ -119,6 +199,34 @@ class TestPage(unittest.TestCase):
         page = Page(**self.page_kwargs)
         self.assertEqual(page.save_as, "pages/foo-bar-fr.html")
 
+    def test_relative_source_path(self):
+        # 'relative_source_path' should be the relative path
+        # from 'PATH' to 'source_path'
+        page_kwargs = self._copy_page_kwargs()
+
+        # If 'source_path' is None, 'relative_source_path' should
+        # also return None
+        page_kwargs['source_path'] = None
+        page = Page(**page_kwargs)
+        self.assertIsNone(page.relative_source_path)
+
+        page_kwargs = self._copy_page_kwargs()
+        settings = get_settings()
+        full_path = page_kwargs['source_path']
+
+        settings['PATH'] = os.path.dirname(full_path)
+        page_kwargs['settings'] = settings
+        page = Page(**page_kwargs)
+
+        # if 'source_path' is set, 'relative_source_path' should
+        # return the relative path from 'PATH' to 'source_path'
+        self.assertEqual(
+            page.relative_source_path,
+            os.path.relpath(
+                full_path,
+                os.path.dirname(full_path)
+            ))
+
     def test_metadata_url_format(self):
         # Arbitrary metadata should be passed through url_format()
         page = Page(**self.page_kwargs)
@@ -129,7 +237,7 @@ class TestPage(unittest.TestCase):
 
     def test_datetime(self):
         # If DATETIME is set to a tuple, it should be used to override LOCALE
-        dt = SafeDatetime(2015, 9, 13)
+        dt = datetime.datetime(2015, 9, 13)
 
         page_kwargs = self._copy_page_kwargs()
 
@@ -138,14 +246,9 @@ class TestPage(unittest.TestCase):
         page = Page(**page_kwargs)
 
         # page.locale_date is a unicode string in both python2 and python3
-        dt_date = dt.strftime(DEFAULT_CONFIG['DEFAULT_DATE_FORMAT']) 
-        # dt_date is a byte string in python2, and a unicode string in python3
-        # Let's make sure it is a unicode string (relies on python 3.3 supporting the u prefix)
-        if type(dt_date) != type(u''):
-            # python2:
-            dt_date = unicode(dt_date, 'utf8')
+        dt_date = dt.strftime(DEFAULT_CONFIG['DEFAULT_DATE_FORMAT'])
 
-        self.assertEqual(page.locale_date, dt_date )
+        self.assertEqual(page.locale_date, dt_date)
         page_kwargs['settings'] = get_settings()
 
         # I doubt this can work on all platforms ...
@@ -180,18 +283,6 @@ class TestPage(unittest.TestCase):
         custom_page = Page(**page_kwargs)
         self.assertEqual('custom', custom_page.template)
 
-    def _copy_page_kwargs(self):
-        # make a deep copy of page_kwargs
-        page_kwargs = dict([(key, self.page_kwargs[key]) for key in
-                            self.page_kwargs])
-        for key in page_kwargs:
-            if not isinstance(page_kwargs[key], dict):
-                break
-            page_kwargs[key] = dict([(subkey, page_kwargs[key][subkey])
-                                     for subkey in page_kwargs[key]])
-
-        return page_kwargs
-
     def test_signal(self):
         def receiver_test_function(sender):
             receiver_test_function.has_been_called = True
@@ -235,15 +326,13 @@ class TestPage(unittest.TestCase):
              '<a href="http://notmyidea.org/category/category.html">link</a>'))
 
     def test_intrasite_link(self):
-        # type does not take unicode in PY2 and bytes in PY3, which in
-        # combination with unicode literals leads to following insane line:
-        cls_name = '_DummyArticle' if six.PY3 else b'_DummyArticle'
+        cls_name = '_DummyArticle'
         article = type(cls_name, (object,), {'url': 'article.html'})
 
         args = self.page_kwargs.copy()
         args['settings'] = get_settings()
         args['source_path'] = 'content'
-        args['context']['filenames'] = {'article.rst': article}
+        args['context']['generated_content'] = {'article.rst': article}
 
         # Classic intrasite link via filename
         args['content'] = (
@@ -297,26 +386,51 @@ class TestPage(unittest.TestCase):
             '?utm_whatever=234&highlight=word#section-2">link</a>'
         )
 
+        # also test for summary in metadata
+        parsed = (
+            'A simple summary test, with a '
+            '<a href="|filename|article.rst">link</a>'
+        )
+        linked = (
+            'A simple summary test, with a '
+            '<a href="http://notmyidea.org/article.html">link</a>'
+        )
+        args['settings']['FORMATTED_FIELDS'] = ['summary', 'custom']
+        args['metadata']['summary'] = parsed
+        args['metadata']['custom'] = parsed
+        args['context']['localsiteurl'] = 'http://notmyidea.org'
+        p = Page(**args)
+        # This is called implicitly from all generators and Pelican.run() once
+        # all files are processed. Here we process just one page so it needs
+        # to be called explicitly.
+        p.refresh_metadata_intersite_links()
+        self.assertEqual(p.summary, linked)
+        self.assertEqual(p.custom, linked)
+
     def test_intrasite_link_more(self):
-        # type does not take unicode in PY2 and bytes in PY3, which in
-        # combination with unicode literals leads to following insane line:
-        cls_name = '_DummyAsset' if six.PY3 else b'_DummyAsset'
+        cls_name = '_DummyAsset'
 
         args = self.page_kwargs.copy()
         args['settings'] = get_settings()
         args['source_path'] = 'content'
-        args['context']['filenames'] = {
-            'images/poster.jpg': type(cls_name, (object,), {'url': 'images/poster.jpg'}),
-            'assets/video.mp4': type(cls_name, (object,), {'url': 'assets/video.mp4'}),
-            'images/graph.svg': type(cls_name, (object,), {'url': 'images/graph.svg'}),
-            'reference.rst': type(cls_name, (object,), {'url': 'reference.html'}),
+        args['context']['static_content'] = {
+            'images/poster.jpg':
+                type(cls_name, (object,), {'url': 'images/poster.jpg'}),
+            'assets/video.mp4':
+                type(cls_name, (object,), {'url': 'assets/video.mp4'}),
+            'images/graph.svg':
+                type(cls_name, (object,), {'url': 'images/graph.svg'}),
+        }
+        args['context']['generated_content'] = {
+            'reference.rst':
+                type(cls_name, (object,), {'url': 'reference.html'}),
         }
 
         # video.poster
         args['content'] = (
             'There is a video with poster '
-            '<video controls poster="{filename}/images/poster.jpg">'
-            '<source src="|filename|/assets/video.mp4" type="video/mp4">'
+            '<video controls poster="{static}/images/poster.jpg">'
+            '<source src="|static|/assets/video.mp4" type="video/mp4">'
             '</video>'
         )
         content = Page(**args).get_content('http://notmyidea.org')
@@ -324,20 +438,25 @@ class TestPage(unittest.TestCase):
             content,
             'There is a video with poster '
             '<video controls poster="http://notmyidea.org/images/poster.jpg">'
-            '<source src="http://notmyidea.org/assets/video.mp4" type="video/mp4">'
+            '<source src="http://notmyidea.org/assets/video.mp4"'
+            ' type="video/mp4">'
             '</video>'
         )
 
         # object.data
         args['content'] = (
             'There is a svg object '
-            '<object data="{filename}/images/graph.svg" type="image/svg+xml"></object>'
+            '<object data="{static}/images/graph.svg"'
+            ' type="image/svg+xml">'
+            '</object>'
         )
         content = Page(**args).get_content('http://notmyidea.org')
         self.assertEqual(
             content,
             'There is a svg object '
-            '<object data="http://notmyidea.org/images/graph.svg" type="image/svg+xml"></object>'
+            '<object data="http://notmyidea.org/images/graph.svg"'
+            ' type="image/svg+xml">'
+            '</object>'
         )
 
         # blockquote.cite
@@ -349,19 +468,122 @@ class TestPage(unittest.TestCase):
         self.assertEqual(
             content,
             'There is a blockquote with cite attribute '
-            '<blockquote cite="http://notmyidea.org/reference.html">blah blah</blockquote>'
+            '<blockquote cite="http://notmyidea.org/reference.html">'
+            'blah blah'
+            '</blockquote>'
         )
 
+    def test_intrasite_link_absolute(self):
+        """Test that absolute URLs are merged properly."""
+
+        args = self.page_kwargs.copy()
+        args['settings'] = get_settings(
+            STATIC_URL='http://static.cool.site/{path}',
+            ARTICLE_URL='http://blog.cool.site/{slug}.html')
+        args['source_path'] = 'content'
+        args['context']['static_content'] = {
+            'images/poster.jpg':
+                Static('', settings=args['settings'],
+                       source_path='images/poster.jpg'),
+        }
+        args['context']['generated_content'] = {
+            'article.rst':
+                Article('', settings=args['settings'], metadata={
+                    'slug': 'article', 'title': 'Article'})
+        }
+
+        # Article link will go to blog
+        args['content'] = (
+            '<a href="{filename}article.rst">Article</a>'
+        )
+        content = Page(**args).get_content('http://cool.site')
+        self.assertEqual(
+            content,
+            '<a href="http://blog.cool.site/article.html">Article</a>'
+        )
+
+        # Page link will go to the main site
+        args['content'] = (
+            '<a href="{index}">Index</a>'
+        )
+        content = Page(**args).get_content('http://cool.site')
+        self.assertEqual(
+            content,
+            '<a href="http://cool.site/index.html">Index</a>'
+        )
+
+        # Image link will go to static
+        args['content'] = (
+            '<img src="{static}/images/poster.jpg"/>'
+        )
+        content = Page(**args).get_content('http://cool.site')
+        self.assertEqual(
+            content,
+            '<img src="http://static.cool.site/images/poster.jpg"/>'
+        )
+
+    def test_intrasite_link_escape(self):
+        article = type(
+            '_DummyArticle', (object,), {'url': 'article-spaces.html'})
+        asset = type(
+            '_DummyAsset', (object,), {'url': 'name@example.com'})
+
+        args = self.page_kwargs.copy()
+        args['settings'] = get_settings()
+        args['source_path'] = 'content'
+        args['context']['generated_content'] = {'article spaces.rst': article}
+        args['context']['static_content'] = {'name@example.com': asset}
+
+        expected_output = (
+            'A simple test with a '
+            '<a href="http://notmyidea.org/article-spaces.html#anchor">link</a> '
+            '<a href="http://notmyidea.org/name@example.com#anchor">file</a>'
+        )
+
+        # not escaped
+        args['content'] = (
+            'A simple test with a '
+            '<a href="{filename}article spaces.rst#anchor">link</a> '
+            '<a href="{static}name@example.com#anchor">file</a>'
+        )
+        content = Page(**args).get_content('http://notmyidea.org')
+        self.assertEqual(content, expected_output)
+
+        # html escaped
+        args['content'] = (
+            'A simple test with a '
+            '<a href="{filename}article spaces.rst#anchor">link</a> '
+            '<a href="{static}name&#64;example.com#anchor">file</a>'
+        )
+        content = Page(**args).get_content('http://notmyidea.org')
+        self.assertEqual(content, expected_output)
+
+        # url escaped
+        args['content'] = (
+            'A simple test with a '
+            '<a href="{filename}article%20spaces.rst#anchor">link</a> '
+            '<a href="{static}name%40example.com#anchor">file</a>'
+        )
+        content = Page(**args).get_content('http://notmyidea.org')
+        self.assertEqual(content, expected_output)
+
+        # html and url escaped
+        args['content'] = (
+            'A simple test with a '
+            '<a href="{filename}article%20spaces.rst#anchor">link</a> '
+            '<a href="{static}name&#64;example.com#anchor">file</a>'
+        )
+        content = Page(**args).get_content('http://notmyidea.org')
+        self.assertEqual(content, expected_output)
+
     def test_intrasite_link_markdown_spaces(self):
-        # Markdown introduces %20 instead of spaces, this tests that
-        # we support markdown doing this.
-        cls_name = '_DummyArticle' if six.PY3 else b'_DummyArticle'
+        cls_name = '_DummyArticle'
         article = type(cls_name, (object,), {'url': 'article-spaces.html'})
 
         args = self.page_kwargs.copy()
         args['settings'] = get_settings()
         args['source_path'] = 'content'
-        args['context']['filenames'] = {'article spaces.rst': article}
+        args['context']['generated_content'] = {'article spaces.rst': article}
 
         # An intrasite link via filename with %20 as a space
         args['content'] = (
@@ -375,6 +597,55 @@ class TestPage(unittest.TestCase):
             '<a href="http://notmyidea.org/article-spaces.html">link</a>'
         )
 
+    def test_intrasite_link_source_and_generated(self):
+        """Test linking both to the source and the generated article
+        """
+        cls_name = '_DummyAsset'
+
+        args = self.page_kwargs.copy()
+        args['settings'] = get_settings()
+        args['source_path'] = 'content'
+        args['context']['generated_content'] = {
+            'article.rst': type(cls_name, (object,), {'url': 'article.html'})}
+        args['context']['static_content'] = {
+            'article.rst': type(cls_name, (object,), {'url': 'article.rst'})}
+
+        args['content'] = (
+            'A simple test, with a link to an'
+            '<a href="{filename}article.rst">article</a> and its'
+            '<a href="{static}article.rst">source</a>'
+        )
+        content = Page(**args).get_content('http://notmyidea.org')
+        self.assertEqual(
+            content,
+            'A simple test, with a link to an'
+            '<a href="http://notmyidea.org/article.html">article</a> and its'
+            '<a href="http://notmyidea.org/article.rst">source</a>'
+        )
+
+    def test_intrasite_link_to_static_content_with_filename(self):
+        """Test linking to a static resource with deprecated {filename}
+        """
+        cls_name = '_DummyAsset'
+
+        args = self.page_kwargs.copy()
+        args['settings'] = get_settings()
+        args['source_path'] = 'content'
+        args['context']['static_content'] = {
+            'poster.jpg':
+                type(cls_name, (object,), {'url': 'images/poster.jpg'})}
+
+        args['content'] = (
+            'A simple test, with a link to a'
+            '<a href="{filename}poster.jpg">poster</a>'
+        )
+        content = Page(**args).get_content('http://notmyidea.org')
+        self.assertEqual(
+            content,
+            'A simple test, with a link to a'
+            '<a href="http://notmyidea.org/images/poster.jpg">poster</a>'
+        )
+
     def test_multiple_authors(self):
         """Test article with multiple authors."""
         args = self.page_kwargs.copy()
@@ -388,7 +659,7 @@ class TestPage(unittest.TestCase):
         assert content.author == content.authors[0]
 
 
-class TestArticle(TestPage):
+class TestArticle(TestBase):
     def test_template(self):
         # Articles default to article, metadata overwrites
         default_article = Article(**self.page_kwargs)
@@ -400,34 +671,102 @@ class TestArticle(TestPage):
 
     def test_slugify_category_author(self):
         settings = get_settings()
-        settings['SLUG_SUBSTITUTIONS'] = [ ('C#', 'csharp') ]
+        settings['SLUG_REGEX_SUBSTITUTIONS'] = [
+            (r'C#', 'csharp'),
+            (r'[^\w\s-]', ''),
+            (r'(?u)\A\s*', ''),
+            (r'(?u)\s*\Z', ''),
+            (r'[-\s]+', '-'),
+        ]
         settings['ARTICLE_URL'] = '{author}/{category}/{slug}/'
         settings['ARTICLE_SAVE_AS'] = '{author}/{category}/{slug}/index.html'
         article_kwargs = self._copy_page_kwargs()
         article_kwargs['metadata']['author'] = Author("O'Brien", settings)
-        article_kwargs['metadata']['category'] = Category('C# & stuff', settings)
+        article_kwargs['metadata']['category'] = Category(
+            'C# & stuff', settings)
         article_kwargs['metadata']['title'] = 'fnord'
         article_kwargs['settings'] = settings
         article = Article(**article_kwargs)
         self.assertEqual(article.url, 'obrien/csharp-stuff/fnord/')
-        self.assertEqual(article.save_as, 'obrien/csharp-stuff/fnord/index.html')
+        self.assertEqual(
+            article.save_as, 'obrien/csharp-stuff/fnord/index.html')
+
+    def test_slugify_with_author_substitutions(self):
+        settings = get_settings()
+        settings['AUTHOR_REGEX_SUBSTITUTIONS'] = [
+            ('Alexander Todorov', 'atodorov'),
+            ('Krasimir Tsonev', 'krasimir'),
+            (r'[^\w\s-]', ''),
+            (r'(?u)\A\s*', ''),
+            (r'(?u)\s*\Z', ''),
+            (r'[-\s]+', '-'),
+        ]
+        settings['ARTICLE_URL'] = 'blog/{author}/{slug}/'
+        settings['ARTICLE_SAVE_AS'] = 'blog/{author}/{slug}/index.html'
+        article_kwargs = self._copy_page_kwargs()
+        article_kwargs['metadata']['author'] = Author('Alexander Todorov',
+                                                      settings)
+        article_kwargs['metadata']['title'] = 'fnord'
+        article_kwargs['settings'] = settings
+        article = Article(**article_kwargs)
+        self.assertEqual(article.url, 'blog/atodorov/fnord/')
+        self.assertEqual(article.save_as, 'blog/atodorov/fnord/index.html')
+
+    def test_slugify_category_with_dots(self):
+        settings = get_settings()
+        settings['CATEGORY_REGEX_SUBSTITUTIONS'] = [
+            ('Fedora QA', 'fedora.qa'),
+        ]
+        settings['ARTICLE_URL'] = '{category}/{slug}/'
+        article_kwargs = self._copy_page_kwargs()
+        article_kwargs['metadata']['category'] = Category('Fedora QA',
+                                                          settings)
+        article_kwargs['metadata']['title'] = 'This Week in Fedora QA'
+        article_kwargs['settings'] = settings
+        article = Article(**article_kwargs)
+        self.assertEqual(article.url, 'fedora.qa/this-week-in-fedora-qa/')
+
+    def test_valid_save_as_detects_breakout(self):
+        settings = get_settings()
+        article_kwargs = self._copy_page_kwargs()
+        article_kwargs['metadata']['slug'] = '../foo'
+        article_kwargs['settings'] = settings
+        article = Article(**article_kwargs)
+        self.assertFalse(article._has_valid_save_as())
+
+    def test_valid_save_as_detects_breakout_to_root(self):
+        settings = get_settings()
+        article_kwargs = self._copy_page_kwargs()
+        article_kwargs['metadata']['slug'] = '/foo'
+        article_kwargs['settings'] = settings
+        article = Article(**article_kwargs)
+        self.assertFalse(article._has_valid_save_as())
+
+    def test_valid_save_as_passes_valid(self):
+        settings = get_settings()
+        article_kwargs = self._copy_page_kwargs()
+        article_kwargs['metadata']['slug'] = 'foo'
+        article_kwargs['settings'] = settings
+        article = Article(**article_kwargs)
+        self.assertTrue(article._has_valid_save_as())
 
 
-class TestStatic(unittest.TestCase):
+class TestStatic(LoggedTestCase):
 
     def setUp(self):
-
+        super().setUp()
         self.settings = get_settings(
             STATIC_SAVE_AS='{path}',
             STATIC_URL='{path}',
             PAGE_SAVE_AS=os.path.join('outpages', '{slug}.html'),
             PAGE_URL='outpages/{slug}.html')
-        self.context = self.settings.copy()
+        self.context = get_context(self.settings)
 
         self.static = Static(content=None, metadata={}, settings=self.settings,
-            source_path=posix_join('dir', 'foo.jpg'), context=self.context)
+                             source_path=posix_join('dir', 'foo.jpg'),
+                             context=self.context)
 
-        self.context['filenames'] = {self.static.source_path: self.static}
+        self.context['static_content'][self.static.source_path] = self.static
 
     def tearDown(self):
         pass
@@ -435,8 +774,10 @@ class TestStatic(unittest.TestCase):
     def test_attach_to_same_dir(self):
         """attach_to() overrides a static file's save_as and url.
         """
-        page = Page(content="fake page",
-            metadata={'title': 'fakepage'}, settings=self.settings,
+        page = Page(
+            content="fake page",
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
             source_path=os.path.join('dir', 'fakepage.md'))
         self.static.attach_to(page)
 
@@ -448,7 +789,7 @@ class TestStatic(unittest.TestCase):
         """attach_to() preserves dirs inside the linking document dir.
         """
         page = Page(content="fake page", metadata={'title': 'fakepage'},
-            settings=self.settings, source_path='fakepage.md')
+                    settings=self.settings, source_path='fakepage.md')
         self.static.attach_to(page)
 
         expected_save_as = os.path.join('outpages', 'dir', 'foo.jpg')
@@ -459,8 +800,8 @@ class TestStatic(unittest.TestCase):
         """attach_to() ignores dirs outside the linking document dir.
         """
         page = Page(content="fake page",
-            metadata={'title': 'fakepage'}, settings=self.settings,
-            source_path=os.path.join('dir', 'otherdir', 'fakepage.md'))
+                    metadata={'title': 'fakepage'}, settings=self.settings,
+                    source_path=os.path.join('dir', 'otherdir', 'fakepage.md'))
         self.static.attach_to(page)
 
         expected_save_as = os.path.join('outpages', 'foo.jpg')
@@ -471,8 +812,8 @@ class TestStatic(unittest.TestCase):
         """attach_to() does nothing when called a second time.
         """
         page = Page(content="fake page",
-            metadata={'title': 'fakepage'}, settings=self.settings,
-            source_path=os.path.join('dir', 'fakepage.md'))
+                    metadata={'title': 'fakepage'}, settings=self.settings,
+                    source_path=os.path.join('dir', 'fakepage.md'))
 
         self.static.attach_to(page)
 
@@ -480,8 +821,10 @@ class TestStatic(unittest.TestCase):
         otherdir_settings.update(dict(
             PAGE_SAVE_AS=os.path.join('otherpages', '{slug}.html'),
             PAGE_URL='otherpages/{slug}.html'))
-        otherdir_page = Page(content="other page",
-            metadata={'title': 'otherpage'}, settings=otherdir_settings,
+        otherdir_page = Page(
+            content="other page",
+            metadata={'title': 'otherpage'},
+            settings=otherdir_settings,
             source_path=os.path.join('dir', 'otherpage.md'))
 
         self.static.attach_to(otherdir_page)
@@ -492,12 +835,14 @@ class TestStatic(unittest.TestCase):
 
     def test_attach_to_does_nothing_after_save_as_referenced(self):
         """attach_to() does nothing if the save_as was already referenced.
-        (For example, by a {filename} link an a document processed earlier.)
+        (For example, by a {static} link an a document processed earlier.)
         """
         original_save_as = self.static.save_as
 
-        page = Page(content="fake page",
-            metadata={'title': 'fakepage'}, settings=self.settings,
+        page = Page(
+            content="fake page",
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
             source_path=os.path.join('dir', 'fakepage.md'))
         self.static.attach_to(page)
 
@@ -506,12 +851,14 @@ class TestStatic(unittest.TestCase):
 
     def test_attach_to_does_nothing_after_url_referenced(self):
         """attach_to() does nothing if the url was already referenced.
-        (For example, by a {filename} link an a document processed earlier.)
+        (For example, by a {static} link an a document processed earlier.)
         """
         original_url = self.static.url
 
-        page = Page(content="fake page",
-            metadata={'title': 'fakepage'}, settings=self.settings,
+        page = Page(
+            content="fake page",
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
             source_path=os.path.join('dir', 'fakepage.md'))
         self.static.attach_to(page)
 
@@ -522,13 +869,15 @@ class TestStatic(unittest.TestCase):
         """attach_to() does not override paths that were overridden elsewhere.
         (For example, by the user with EXTRA_PATH_METADATA)
         """
-        customstatic = Static(content=None,
+        customstatic = Static(
+            content=None,
             metadata=dict(save_as='customfoo.jpg', url='customfoo.jpg'),
             settings=self.settings,
             source_path=os.path.join('dir', 'foo.jpg'),
             context=self.settings.copy())
 
-        page = Page(content="fake page",
+        page = Page(
+            content="fake page",
             metadata={'title': 'fakepage'}, settings=self.settings,
             source_path=os.path.join('dir', 'fakepage.md'))
 
@@ -541,13 +890,16 @@ class TestStatic(unittest.TestCase):
         """{attach} link syntax triggers output path override & url replacement.
         """
         html = '<a href="{attach}../foo.jpg">link</a>'
-        page = Page(content=html,
-            metadata={'title': 'fakepage'}, settings=self.settings,
+        page = Page(
+            content=html,
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
             source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
             context=self.context)
         content = page.get_content('')
 
-        self.assertNotEqual(content, html,
+        self.assertNotEqual(
+            content, html,
             "{attach} link syntax did not trigger URL replacement.")
 
         expected_save_as = os.path.join('outpages', 'foo.jpg')
@@ -560,7 +912,8 @@ class TestStatic(unittest.TestCase):
         html = '<a href="{tag}foo">link</a>'
         page = Page(
             content=html,
-            metadata={'title': 'fakepage'}, settings=self.settings,
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
             source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
             context=self.context)
         content = page.get_content('')
@@ -571,37 +924,115 @@ class TestStatic(unittest.TestCase):
         "{category} link syntax triggers url replacement."
 
         html = '<a href="{category}foo">link</a>'
-        page = Page(content=html,
-            metadata={'title': 'fakepage'}, settings=self.settings,
+        page = Page(
+            content=html,
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
             source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
             context=self.context)
         content = page.get_content('')
 
         self.assertNotEqual(content, html)
 
+    def test_author_link_syntax(self):
+        "{author} link syntax triggers url replacement."
 
-class TestURLWrapper(unittest.TestCase):
-    def test_comparisons(self):
-        # URLWrappers are sorted by name
-        wrapper_a = URLWrapper(name='first', settings={})
-        wrapper_b = URLWrapper(name='last', settings={})
-        self.assertFalse(wrapper_a > wrapper_b)
-        self.assertFalse(wrapper_a >= wrapper_b)
-        self.assertFalse(wrapper_a == wrapper_b)
-        self.assertTrue(wrapper_a != wrapper_b)
-        self.assertTrue(wrapper_a <= wrapper_b)
-        self.assertTrue(wrapper_a < wrapper_b)
-        wrapper_b.name = 'first'
-        self.assertFalse(wrapper_a > wrapper_b)
-        self.assertTrue(wrapper_a >= wrapper_b)
-        self.assertTrue(wrapper_a == wrapper_b)
-        self.assertFalse(wrapper_a != wrapper_b)
-        self.assertTrue(wrapper_a <= wrapper_b)
-        self.assertFalse(wrapper_a < wrapper_b)
-        wrapper_a.name = 'last'
-        self.assertTrue(wrapper_a > wrapper_b)
-        self.assertTrue(wrapper_a >= wrapper_b)
-        self.assertFalse(wrapper_a == wrapper_b)
-        self.assertTrue(wrapper_a != wrapper_b)
-        self.assertFalse(wrapper_a <= wrapper_b)
-        self.assertFalse(wrapper_a < wrapper_b)
+        html = '<a href="{author}foo">link</a>'
+        page = Page(
+            content=html,
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
+            source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
+            context=self.context)
+        content = page.get_content('')
+
+        self.assertNotEqual(content, html)
+
+    def test_index_link_syntax(self):
+        "{index} link syntax triggers url replacement."
+
+        html = '<a href="{index}">link</a>'
+        page = Page(
+            content=html,
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
+            source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
+            context=self.context)
+        content = page.get_content('')
+
+        self.assertNotEqual(content, html)
+
+        expected_html = ('<a href="' +
+                         '/'.join((self.settings['SITEURL'],
+                                   self.settings['INDEX_SAVE_AS'])) +
+                         '">link</a>')
+        self.assertEqual(content, expected_html)
+
+    def test_unknown_link_syntax(self):
+        "{unknown} link syntax should trigger warning."
+
+        html = '<a href="{unknown}foo">link</a>'
+        page = Page(content=html,
+                    metadata={'title': 'fakepage'}, settings=self.settings,
+                    source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
+                    context=self.context)
+        content = page.get_content('')
+
+        self.assertEqual(content, html)
+        self.assertLogCountEqual(
+            count=1,
+            msg="Replacement Indicator 'unknown' not recognized, "
+                "skipping replacement",
+            level=logging.WARNING)
+
+    def test_link_to_unknown_file(self):
+        "{filename} link to unknown file should trigger warning."
+
+        html = '<a href="{filename}foo">link</a>'
+        page = Page(content=html,
+                    metadata={'title': 'fakepage'}, settings=self.settings,
+                    source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
+                    context=self.context)
+        content = page.get_content('')
+
+        self.assertEqual(content, html)
+        self.assertLogCountEqual(
+            count=1,
+            msg="Unable to find 'foo', skipping url replacement.",
+            level=logging.WARNING)
+
+    def test_index_link_syntax_with_spaces(self):
+        """{index} link syntax triggers url replacement
+        with spaces around the equal sign."""
+
+        html = '<a href = "{index}">link</a>'
+        page = Page(
+            content=html,
+            metadata={'title': 'fakepage'},
+            settings=self.settings,
+            source_path=os.path.join('dir', 'otherdir', 'fakepage.md'),
+            context=self.context)
+        content = page.get_content('')
+
+        self.assertNotEqual(content, html)
+
+        expected_html = ('<a href = "' +
+                         '/'.join((self.settings['SITEURL'],
+                                   self.settings['INDEX_SAVE_AS'])) +
+                         '">link</a>')
+        self.assertEqual(content, expected_html)
+
+    def test_not_save_as_draft(self):
+        """Static.save_as is not affected by draft status."""
+
+        static = Static(
+            content=None,
+            metadata=dict(status='draft',),
+            settings=self.settings,
+            source_path=os.path.join('dir', 'foo.jpg'),
+            context=self.settings.copy())
+
+        expected_save_as = posixize_path(os.path.join('dir', 'foo.jpg'))
+        self.assertEqual(static.status, 'draft')
+        self.assertEqual(static.save_as, expected_save_as)
+        self.assertEqual(static.url, path_to_url(expected_save_as))
