@@ -1,6 +1,4 @@
 #!/usr/bin/env python
-# -*- coding: utf-8 -*-
-from __future__ import print_function, unicode_literals
 
 import argparse
 import logging
@@ -8,23 +6,18 @@ import os
 import re
 import subprocess
 import sys
-from codecs import open
+import time
 from collections import defaultdict
-
-from six.moves.urllib.error import URLError
-from six.moves.urllib.parse import urlparse
-from six.moves.urllib.request import urlretrieve
+from html import unescape
+from urllib.error import URLError
+from urllib.parse import quote, urlparse, urlsplit, urlunsplit
+from urllib.request import urlretrieve
 
 # because logging.setLoggerClass has to be called before logging.getLogger
 from pelican.log import init
 from pelican.settings import read_settings
 from pelican.utils import SafeDatetime, slugify
 
-try:
-    from html import unescape  # py3.4+
-except ImportError:
-    from six.moves.html_parser import HTMLParser
-    unescape = HTMLParser().unescape
 
 logger = logging.getLogger(__name__)
 
@@ -46,7 +39,7 @@ def decode_wp_content(content, br=True):
             if start == -1:
                 content = content + pre_part
                 continue
-            name = "<pre wp-pre-tag-{0}></pre>".format(pre_index)
+            name = "<pre wp-pre-tag-{}></pre>".format(pre_index)
             pre_tags[name] = pre_part[start:] + "</pre>"
             content = content + pre_part[0:start] + name
             pre_index += 1
@@ -131,11 +124,11 @@ def xml_to_soup(xml):
     return soup
 
 
-def get_filename(filename, post_id):
-    if filename is not None:
-        return filename
-    else:
+def get_filename(post_name, post_id):
+    if post_name is None or post_name.isspace():
         return post_id
+    else:
+        return post_name
 
 
 def wp2fields(xml, wp_custpost=False):
@@ -154,13 +147,13 @@ def wp2fields(xml, wp_custpost=False):
                 title = 'No title [%s]' % item.find('post_name').string
                 logger.warning('Post "%s" is lacking a proper title', title)
 
-            filename = item.find('post_name').string
+            post_name = item.find('post_name').string
             post_id = item.find('post_id').string
-            filename = get_filename(filename, post_id)
+            filename = get_filename(post_name, post_id)
 
             content = item.find('encoded').string
             raw_date = item.find('post_date').string
-            if raw_date == u'0000-00-00 00:00:00':
+            if raw_date == '0000-00-00 00:00:00':
                 date = None
             else:
                 date_object = SafeDatetime.strptime(
@@ -264,7 +257,7 @@ def dc2fields(file):
     category_list = {}
     posts = []
 
-    with open(file, 'r', encoding='utf-8') as f:
+    with open(file, encoding='utf-8') as f:
 
         for line in f:
             # remove final \n
@@ -395,23 +388,12 @@ def posterous2fields(api_token, email, password):
     """Imports posterous posts"""
     import base64
     from datetime import timedelta
-    try:
-        # py3k import
-        import json
-    except ImportError:
-        # py2 import
-        import simplejson as json
-
-    try:
-        # py3k import
-        import urllib.request as urllib_request
-    except ImportError:
-        # py2 import
-        import urllib2 as urllib_request
+    import json
+    import urllib.request as urllib_request
 
     def get_posterous_posts(api_token, email, password, page=1):
         base64string = base64.encodestring(
-            ("%s:%s" % (email, password)).encode('utf-8')).replace('\n', '')
+            ("{}:{}".format(email, password)).encode('utf-8')).replace('\n', '')
         url = ("http://posterous.com/api/v2/users/me/sites/primary/"
                "posts?api_token=%s&page=%d") % (api_token, page)
         request = urllib_request.Request(url)
@@ -450,22 +432,11 @@ def posterous2fields(api_token, email, password):
 
 def tumblr2fields(api_key, blogname):
     """ Imports Tumblr posts (API v2)"""
-    try:
-        # py3k import
-        import json
-    except ImportError:
-        # py2 import
-        import simplejson as json
-
-    try:
-        # py3k import
-        import urllib.request as urllib_request
-    except ImportError:
-        # py2 import
-        import urllib2 as urllib_request
+    import json
+    import urllib.request as urllib_request
 
     def get_tumblr_posts(api_key, blogname, offset=0):
-        url = ("http://api.tumblr.com/v2/blog/%s.tumblr.com/"
+        url = ("https://api.tumblr.com/v2/blog/%s.tumblr.com/"
                "posts?api_key=%s&offset=%d&filter=raw") % (
             blogname, api_key, offset)
         request = urllib_request.Request(url)
@@ -563,7 +534,7 @@ def feed2fields(file):
     settings = read_settings()
     subs = settings['SLUG_REGEX_SUBSTITUTIONS']
     for entry in d.entries:
-        date = (entry.updated_parsed.strftime('%Y-%m-%d %H:%M')
+        date = (time.strftime('%Y-%m-%d %H:%M', entry.updated_parsed)
                 if hasattr(entry, 'updated_parsed') else None)
         author = entry.author if hasattr(entry, 'author') else None
         tags = ([e['term'] for e in entry.tags]
@@ -581,7 +552,7 @@ def build_header(title, date, author, categories, tags, slug,
 
     from docutils.utils import column_width
 
-    header = '%s\n%s\n' % (title, '#' * column_width(title))
+    header = '{}\n{}\n'.format(title, '#' * column_width(title))
     if date:
         header += ':date: %s\n' % date
     if author:
@@ -600,6 +571,29 @@ def build_header(title, date, author, categories, tags, slug,
     return header
 
 
+def build_asciidoc_header(title, date, author, categories, tags, slug,
+                          status=None, attachments=None):
+    """Build a header from a list of fields"""
+
+    header = '= %s\n' % title
+    if author:
+        header += '%s\n' % author
+        if date:
+            header += '%s\n' % date
+    if categories:
+        header += ':category: %s\n' % ', '.join(categories)
+    if tags:
+        header += ':tags: %s\n' % ', '.join(tags)
+    if slug:
+        header += ':slug: %s\n' % slug
+    if status:
+        header += ':status: %s\n' % status
+    if attachments:
+        header += ':attachments: %s\n' % ', '.join(attachments)
+    header += '\n'
+    return header
+
+
 def build_markdown_header(title, date, author, categories, tags,
                           slug, status=None, attachments=None):
     """Build a header from a list of fields"""
@@ -623,7 +617,9 @@ def build_markdown_header(title, date, author, categories, tags,
 
 
 def get_ext(out_markup, in_markup='html'):
-    if in_markup == 'markdown' or out_markup == 'markdown':
+    if out_markup == 'asciidoc':
+        ext = '.adoc'
+    elif in_markup == 'markdown' or out_markup == 'markdown':
         ext = '.md'
     else:
         ext = '.rst'
@@ -635,7 +631,7 @@ def get_out_filename(output_path, filename, ext, kind,
     filename = os.path.basename(filename)
 
     # Enforce filename restrictions for various filesystems at once; see
-    # http://en.wikipedia.org/wiki/Filename#Reserved_characters_and_words
+    # https://en.wikipedia.org/wiki/Filename#Reserved_characters_and_words
     # we do not need to filter words because an extension will be appended
     filename = re.sub(r'[<>:"/\\|?*^% ]', '-', filename)  # invalid chars
     filename = filename.lstrip('.')  # should not start with a dot
@@ -662,7 +658,8 @@ def get_out_filename(output_path, filename, ext, kind,
             typename = ''
             kind = 'article'
         if dircat and (len(categories) > 0):
-            catname = slugify(categories[0], regex_subs=slug_subs)
+            catname = slugify(
+                categories[0], regex_subs=slug_subs, preserve_case=True)
         else:
             catname = ''
         out_filename = os.path.join(output_path, typename,
@@ -671,7 +668,8 @@ def get_out_filename(output_path, filename, ext, kind,
             os.makedirs(os.path.join(output_path, typename, catname))
     # option to put files in directories with categories names
     elif dircat and (len(categories) > 0):
-        catname = slugify(categories[0], regex_subs=slug_subs)
+        catname = slugify(
+            categories[0], regex_subs=slug_subs, preserve_case=True)
         out_filename = os.path.join(output_path, catname, filename + ext)
         if not os.path.isdir(os.path.join(output_path, catname)):
             os.mkdir(os.path.join(output_path, catname))
@@ -690,16 +688,16 @@ def get_attachments(xml):
 
     for item in items:
         kind = item.find('post_type').string
-        filename = item.find('post_name').string
+        post_name = item.find('post_name').string
         post_id = item.find('post_id').string
 
         if kind == 'attachment':
             attachments.append((item.find('post_parent').string,
                                 item.find('attachment_url').string))
         else:
-            filename = get_filename(filename, post_id)
+            filename = get_filename(post_name, post_id)
             names[post_id] = filename
-    attachedposts = defaultdict(list)
+    attachedposts = defaultdict(set)
     for parent, url in attachments:
         try:
             parent_name = names[parent]
@@ -707,7 +705,7 @@ def get_attachments(xml):
             # attachment's parent is not a valid post
             parent_name = None
 
-        attachedposts[parent_name].append(url)
+        attachedposts[parent_name].add(url)
     return attachedposts
 
 
@@ -727,21 +725,27 @@ def download_attachments(output_path, urls):
             if sys.platform != 'win32' or ':' not in item:
                 localpath = os.path.join(localpath, item)
         full_path = os.path.join(output_path, localpath)
+
+        # Generate percent-encoded URL
+        scheme, netloc, path, query, fragment = urlsplit(url)
+        if scheme != 'file':
+            path = quote(path)
+            url = urlunsplit((scheme, netloc, path, query, fragment))
+
         if not os.path.exists(full_path):
             os.makedirs(full_path)
         print('downloading {}'.format(filename))
         try:
             urlretrieve(url, os.path.join(full_path, filename))
             locations[url] = os.path.join(localpath, filename)
-        except (URLError, IOError) as e:
+        except (URLError, OSError) as e:
             # Python 2.7 throws an IOError rather Than URLError
             logger.warning("No file could be downloaded from %s\n%s", url, e)
     return locations
 
 
-def is_pandoc_needed(fields):
-    in_markup_idx = 9
-    return filter(lambda f: f[in_markup_idx] in ('html', 'wp-html'), fields)
+def is_pandoc_needed(in_markup):
+    return in_markup in ('html', 'wp-html')
 
 
 def get_pandoc_version():
@@ -772,11 +776,7 @@ def fields2pelican(
         wp_custpost=False, wp_attach=False, attachments=None):
 
     pandoc_version = get_pandoc_version()
-
-    if is_pandoc_needed(fields) and not pandoc_version:
-        error = ('Pandoc must be installed to complete the '
-                 'requested import action.')
-        exit(error)
+    posts_require_pandoc = []
 
     settings = read_settings()
     slug_subs = settings['SLUG_REGEX_SUBSTITUTIONS']
@@ -785,6 +785,9 @@ def fields2pelican(
             kind, in_markup) in fields:
         if filter_author and filter_author != author:
             continue
+        if is_pandoc_needed(in_markup) and not pandoc_version:
+            posts_require_pandoc.append(filename)
+
         slug = not disable_slugs and filename or None
 
         if wp_attach and attachments:
@@ -797,7 +800,10 @@ def fields2pelican(
             links = None
 
         ext = get_ext(out_markup, in_markup)
-        if ext == '.md':
+        if ext == '.adoc':
+            header = build_asciidoc_header(title, date, author, categories,
+                                           tags, slug, status, attachments)
+        elif ext == '.md':
             header = build_markdown_header(
                 title, date, author, categories, tags, slug,
                 status, links.values() if links else None)
@@ -822,7 +828,7 @@ def fields2pelican(
                     new_content = decode_wp_content(content)
                 else:
                     paragraphs = content.splitlines()
-                    paragraphs = ['<p>{0}</p>'.format(p) for p in paragraphs]
+                    paragraphs = ['<p>{}</p>'.format(p) for p in paragraphs]
                     new_content = ''.join(paragraphs)
 
                 fp.write(new_content)
@@ -856,7 +862,7 @@ def fields2pelican(
 
             os.remove(html_filename)
 
-            with open(out_filename, 'r', encoding='utf-8') as fs:
+            with open(out_filename, encoding='utf-8') as fs:
                 content = fs.read()
                 if out_markup == 'markdown':
                     # In markdown, to insert a <br />, end a line with two
@@ -869,6 +875,11 @@ def fields2pelican(
 
         with open(out_filename, 'w', encoding='utf-8') as fs:
             fs.write(header + content)
+
+    if posts_require_pandoc:
+        logger.error("Pandoc must be installed to import the following posts:"
+                     "\n  {}".format("\n  ".join(posts_require_pandoc)))
+
     if wp_attach and attachments and None in attachments:
         print("downloading attachments that don't have a parent post")
         urls = attachments[None]
@@ -877,7 +888,7 @@ def fields2pelican(
 
 def main():
     parser = argparse.ArgumentParser(
-        description="Transform feed, Blogger, Dotclear, Posterous, Tumblr, or"
+        description="Transform feed, Blogger, Dotclear, Posterous, Tumblr, or "
                     "WordPress files into reST (rst) or Markdown (md) files. "
                     "Be sure to have pandoc installed.",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
